/// ReMixer
/// A Public Domain C++ Audio Playback Library
/// By william @ RedactedSoftware. Thanks to Dawsh & Maxine.
/// (c) 2024 Redacted Software
/// This work is explicitly dedicated to the public domain, for the hopeful betterment of the software industry.

/// @file pacat_simple.cpp
/// @desc A CLI program that plays back raw PCM data (given as a file) over the PulseAudio service.
/// @edit 2024-08-19

#include <iostream>

#include <cstdio>
#include <unistd.h>
#include <cstring>
#include <cerrno>
#include <fcntl.h>
#include <pulse/simple.h>
#include <pulse/error.h>
#include <pulse/volume.h>
#include <pulse/thread-mainloop.h>
#define BUFSIZE 1024

int main(int argc, char* argv[])
{
    if (argc == 1 || argv[1] == "-h" || argv[1] == "help")
    {
        std::cout << "pacat - Tests playback of raw audio data via PulseAudio API." << std::endl;
        std::cout << "usage: pacat <file> (presumably output.raw)\n" << std::endl;
    }

    static const pa_sample_spec ss = {
            .format = PA_SAMPLE_S16LE,
            .rate = 44100,
            .channels = 2
    };

    pa_simple *s = NULL;

    int ret = 1;
    int error;

    /* Replace STDIN with the specified file if needed. */
    if (argc > 1) {
        int fd;
        if ((fd = open(argv[1], O_RDONLY)) < 0) {
            fprintf(stderr, __FILE__": open() failed: %s\n", strerror(errno));
            goto finish;
        }

        if (dup2(fd, STDIN_FILENO) < 0) {
            fprintf(stderr, __FILE__": dup2() failed: %s\n", strerror(errno));
            goto finish;
        }
        close(fd);
    }

    /* Create a new playback stream */
    if (!(s = pa_simple_new(NULL, argv[0], PA_STREAM_PLAYBACK, NULL, "playback", &ss, NULL, NULL, &error))) {
        fprintf(stderr, __FILE__": pa_simple_new() failed: %s\n", pa_strerror(error));
        goto finish;
    }

    for (;;) {
        uint8_t buf[BUFSIZE];
        ssize_t r;

#if 0
        pa_usec_t latency;

            if ((latency = pa_simple_get_latency(s, &error)) == (pa_usec_t)-1)
            {
                fprintf(stderr, __FILE__": pa_simple_get_latency() failed: %s\n", pa_strerror(error));
                return;
            }
            fprintf(stderr, "%0.0f usec     \r", (float)latency);
#endif
        /* Read some data */
        if ((r = read(STDIN_FILENO, buf, sizeof(buf))) <= 0) {
            if (r == 0) /* EOF */
                break;

            fprintf(stderr, __FILE__": read() failed: %s\n", strerror(errno));
            goto finish;
        }

        /* and play it */
        if (pa_simple_write(s, buf, (size_t)r, &error) < 0) {
            fprintf(stderr, __FILE__": pa_simple_write() failed: %s\n", pa_strerror(error));
            goto finish;
        }
    }

    /* Make sure that every single sample was played */
    if (pa_simple_drain(s, &error) < 0) {
        fprintf(stderr, __FILE__": pa_simple_drain() failed: %s\n", pa_strerror(error));
        goto finish;
    }

    ret = 0;
    finish:
    if (s)
        pa_simple_free(s);

    return ret;
}
