#include <../../include/Backend/Pulse/PulseDevice.h>

void ReMixer::PulseAudioSubsystem::Play(ReMixer::PulseStream &stream, const ReMixer::Sound &sound) {

    // Check parameters and compatibility between stream and sound.
    if (stream.SampleRate() != sound.SampleRate()) {
        std::cerr << "There is an incongruency in sample rate between the stream and the sound file. "
                  << std::format("Stream: {}hz, meanwhile sound: {}hz", stream.SampleRate(), sound.SampleRate())
                  << std::endl;

        stream.SetSampleRate(sound.SampleRate());
    }
    pa_threaded_mainloop_lock(pa->mainloop);

    size_t length = sound.SampleCount();

    const void* data = sound.ptr();

    while(length > 0)
    {
        size_t l;
        int r;
        while(!((l = pa_stream_writable_size(stream.stream)))) {
            pa_threaded_mainloop_wait(pa->mainloop);
        }

        if (l > length)
            l = length;

        r = pa_stream_write(stream.stream, data, l, nullptr, 0LL, PA_SEEK_RELATIVE);
        data = (const char*) data + l;
        length -= l;
    }

    pa_threaded_mainloop_unlock(pa->mainloop);
}

bool ReMixer::PulseAudioSubsystem::Connected() const {
    return ctx_connected;
}

std::string ReMixer::PulseAudioSubsystem::ContextName() const {
    return ctx_name;
}

std::string ReMixer::PulseAudioSubsystem::ServerName() const
{
    return pa_context_get_server(this->pa->context);
}

void ReMixer::PulseAudioSubsystem::SetContextName(const std::string &new_context_name) {
    this->ctx_name = new_context_name;
    pa_context_set_name(this->pa->context, new_context_name.c_str(), OnCtxSuccess, this);
}

ReMixer::PulseStream ReMixer::PulseAudioSubsystem::CreateStream(const std::string& stream_name) {
    // Assumed defaults
    // TODO: Expose to the API later
    static const pa_sample_spec ss = {
            .format = PA_SAMPLE_S16LE,
            .rate = 44.1_kHz,
            .channels = 2
    };

    const pa_cvolume* volume = nullptr;
    pa_channel_map* map = nullptr;
    const char* dev = nullptr;
    pa_buffer_attr* attr = nullptr;

    PulseStream new_stream = PulseStream();
    new_stream.SetName(stream_name);
    new_stream.sample_rate = 44.1_kHz;
    pa_threaded_mainloop_lock(pa->mainloop);

    if (!(new_stream.stream = pa_stream_new(pa->context, stream_name.c_str(), &ss, map)))
    {
        throw std::runtime_error("Failed to create a new PulseAudio stream.");
    }

    pa_stream_set_state_callback(new_stream.stream, PulseStream::OnStreamStateChanged, pa->mainloop);
    pa_stream_set_read_callback(new_stream.stream, PulseStream::OnStreamRequest, pa->mainloop);
    pa_stream_set_write_callback(new_stream.stream, PulseStream::OnStreamRequest, pa->mainloop);
    pa_stream_set_latency_update_callback(new_stream.stream, PulseStream::OnLatencyUpdate, pa->mainloop);

    pa_stream_flags flags = static_cast<pa_stream_flags>(PA_STREAM_INTERPOLATE_TIMING | PA_STREAM_ADJUST_LATENCY |
                                                         PA_STREAM_AUTO_TIMING_UPDATE | PA_STREAM_VARIABLE_RATE);

    // TODO: Support Playback and Recording streams
    int r = pa_stream_connect_playback(new_stream.stream, dev, attr, flags, volume, NULL);
    if (r < 0)
    {
        int error = pa_context_errno(pa->context);
        throw std::runtime_error(std::format("Failed to connect the stream: {}", error));
    }

    for (;;)
    {
        pa_stream_state_t state;
        state = pa_stream_get_state(new_stream.stream);

        if (state == PA_STREAM_READY)
            break;

        if (!PA_STREAM_IS_GOOD(state))
        {
            int error = pa_context_errno(pa->context);
            throw std::runtime_error(std::format("Stream state is bad: {}", error));
        }

        pa_threaded_mainloop_wait(pa->mainloop);
    }

    pa_threaded_mainloop_unlock(pa->mainloop);

    streams.insert({stream_name, new_stream});

    return new_stream;
}

ReMixer::PulseAudioSubsystem::PulseAudioSubsystem(const std::string &context_name) : SoundSubsystem()
{
    ctx_name = context_name;
    pa = pa_xnew(pa_impl, 1);

    const char *server = NULL;
    int error = PA_ERR_INTERNAL;

    // Attempt to create a self-threaded PulseAudio instance.
    if (!(pa->mainloop = pa_threaded_mainloop_new())) {
        throw std::runtime_error("Failed to create a PulseAudio mainloop.");
    }

    // Grab our API handle.
    pa->api = pa_threaded_mainloop_get_api(pa->mainloop);

    // Attempt to setup our context.
    if (!(pa->context = pa_context_new(pa->api, context_name.c_str()))) {
        throw std::runtime_error("Failed to create a PulseAudio context.");
    }

    pa_context_flags flags = pa_context_flags::PA_CONTEXT_NOAUTOSPAWN;


    pa_context_set_state_callback(pa->context, OnCtxStateChange, this);

    // Attempt to connect to the PulseAudio server.
    if (pa_context_connect(pa->context, server, flags, NULL) < 0)
    {
        error = pa_context_errno(pa->context);
        throw std::runtime_error(std::format("Failed to connect to context {}", error));
    } else{
        ctx_connected = true;
    }

    pa_threaded_mainloop_lock(pa->mainloop);

    // Attempt to Start the PulseAudio Loop.
    if (pa_threaded_mainloop_start(pa->mainloop) < 0) {
        throw std::runtime_error("Failed to start thread mainloop.");
    }

    // Wait until the context is ready, listen for updates from the server.
    for (;;) {
        pa_context_state_t state;
        state = pa_context_get_state(pa->context);

        if (state == PA_CONTEXT_READY)
            break;

        if (!PA_CONTEXT_IS_GOOD(state)) {
            error = pa_context_errno(pa->context);

            throw std::runtime_error(std::format("Context was not good: {}", error));
        }

        pa_threaded_mainloop_wait(pa->mainloop);
    }


    pa_threaded_mainloop_unlock(pa->mainloop);

    pa_context_get_server_info(pa->context, OnListServerInfo, this);
    pa_context_get_sink_info_list(pa->context, OnListSinkInfo, this);

    std::cout << "PulseAudio context is ready." << std::endl;
}

void ReMixer::PulseAudioSubsystem::OnCtxSuccess(pa_context* c, int success, void* userdata)
{
    auto* p = static_cast<PulseAudioSubsystem *>(userdata);

    if (success)
        p->OnSuccess.Invoke(true);
    else
        p->OnSuccess.Invoke(false);
}

void ReMixer::PulseAudioSubsystem::OnCtxStateChange(pa_context *c, void *userdata) {
    std::cout << "context state callback: ";

    auto* p = static_cast<PulseAudioSubsystem *>(userdata);


    switch(pa_context_get_state(c)) {
        case PA_CONTEXT_READY:
            std::cout << "Context Ready" << std::endl;
            pa_threaded_mainloop_signal(p->pa->mainloop, 0);
            break;
        case PA_CONTEXT_TERMINATED:
            std::cout << "Context Terminated" << std::endl;
            pa_threaded_mainloop_signal(p->pa->mainloop, 0);
            break;
        case PA_CONTEXT_FAILED:
            std::cout << "Context Failed" << std::endl;
            pa_threaded_mainloop_signal(p->pa->mainloop, 0);
            break;
        case PA_CONTEXT_UNCONNECTED:
            std::cout << "Context Unconnected" << std::endl;
            break;
        case PA_CONTEXT_CONNECTING:
            std::cout << "Context Connecting" << std::endl;
            break;
        case PA_CONTEXT_AUTHORIZING:
            std::cout << "Context Authorizing" << std::endl;
            break;
        case PA_CONTEXT_SETTING_NAME:
            std::cout << "Context Setting Name" << std::endl;
            break;
        default:
            std::cerr << "The fuck?" << std::endl;
            break;
    }
}

void ReMixer::PulseAudioSubsystem::ListDevices() {
    pa_threaded_mainloop_lock(pa->mainloop);

    pa_operation * op;
    // List 'Sinks' aka playback devices
    op = pa_context_get_sink_info_list(pa->context, OnSinkDevice, this);

    for (;;)
    {
        int r = pa_operation_get_state(op);

        if (r == PA_OPERATION_DONE || r == PA_OPERATION_CANCELLED)
            break;

        pa_threaded_mainloop_wait(pa->mainloop);
    }

    pa_operation_unref(op);

    pa_threaded_mainloop_unlock(pa->mainloop);
}

void ReMixer::PulseAudioSubsystem::OnListServerInfo(pa_context *c, const pa_server_info *i, void *userdata) {
    auto* p = static_cast<PulseAudioSubsystem*>(userdata);

    p->OnServerInfo.Invoke(i);

    pa_threaded_mainloop_signal(p->pa->mainloop, 0);
}

void ReMixer::PulseAudioSubsystem::OnListSinkInfo(pa_context *c, const pa_sink_info *i, int eol, void *userdata) {

}

void ReMixer::PulseAudioSubsystem::OnSinkDevice(pa_context* c, const pa_sink_info* info, int eol, void* userdata)
{
    if (eol != 0)
    {
        auto* p = static_cast<PulseAudioSubsystem*> (userdata);

        pa_threaded_mainloop_signal(p->pa->mainloop, 0);
    }

    if (info)
    {

    }
}