#include <ReMixer/Sound.hpp>
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>
#include <fstream>
#include <format>
#include <cstring>


ReMixer::Sound ReMixer::Sound::FromOGGVorbisFile(const std::filesystem::path &file_name) {
    Sound sound;
    sound.LoadOGGVorbisFile(file_name);
    return sound;
}

ReMixer::Sound ReMixer::Sound::FromWAVFile(const std::filesystem::path &file_name) {
    Sound sound;
    sound.LoadWAVFile(file_name);
    return sound;
}

ReMixer::Sound ReMixer::Sound::FromPCMFile(const std::filesystem::path &file_name, uint sample_rate, StreamMode stream_mode) {
    Sound sound;
    sound.LoadPCMFile(file_name, sample_rate, stream_mode);
    return sound;
}

ReMixer::Sound ReMixer::Sound::FromPCMBuffer(std::vector<char> buffer, uint sample_rate, StreamMode stream_mode) {
    Sound sound;
    sound.LoadPCMBuffer(buffer, sample_rate, stream_mode);
    return sound;
}

void ReMixer::Sound::LoadPCMFile(const std::filesystem::path &file_name, uint sampleRate, StreamMode streamMode) {
    std::ifstream file(file_name, std::ios::binary | std::ios::ate);

    if (!file.is_open())
    {
        throw std::runtime_error(std::format("No such file {} could be found", file_name.c_str()));
    }

    file.unsetf(std::ios::skipws);

    file.seekg(0, std::ios::end);
    std::streamsize size = file.tellg();
    file.seekg(0, std::ios::beg);

    audio_data.reserve(size);

    audio_data.insert(audio_data.begin(),
                      std::istream_iterator<char>(file),
                      std::istream_iterator<char>());

    this->sample_rate = sampleRate;
    this->stream_mode = streamMode;
}

void ReMixer::Sound::LoadPCMBuffer(std::vector<char> buffer, uint sampleRate, StreamMode streamMode) {
    audio_data = buffer;
    this->sample_rate = sampleRate;
    this->stream_mode = streamMode;
}

void ReMixer::Sound::LoadOGGVorbisFile(const std::filesystem::path &file_name) {
    OggVorbis_File vf;
    FILE* inFile = fopen(file_name.c_str(), "rb");
    if (!inFile) {
        std::cerr << "Error opening input file." << std::endl;
        return;
    }

    if (ov_open(inFile, &vf, NULL, 0) < 0) {
        std::cerr << "Error opening Ogg Vorbis file." << std::endl;
        fclose(inFile);
        return;
    }

    vorbis_info* vi = ov_info(&vf, -1);

    if (vi->channels > 2)
    {
        std::cerr << "Warning: Our sound system currently only supports Mono and Stereo. "
                  << std::format("However, we detect {} number of channels in this sound file. ", vi->channels)
                  << "Information may be lost, but we will do our best to play it back anyway." << std::endl;


        stream_mode = StreamMode::STEREO;
    } else if (vi->channels == 2) {
        stream_mode = StreamMode::STEREO;
    } else if (vi->channels == 1) {
        stream_mode = StreamMode::MONO;
    } else if (vi->channels == 0) {
        throw std::runtime_error("We don't know what to do with this file?");
    }

    sample_rate = vi->rate;
    //vi->

    //std::cout << vi->bitrate_nominal << std::endl;
    //std::cout << vi->bitrate_lower << std::endl;
    //std::cout << vi->bitrate_upper << std::endl;
    //std::cout << vi->bitrate_window << std::endl;
    //std::cout << vi->rate << std::endl;
    //std::cout << vi->version << std::endl;
    //std::cout << vi->channels << std::endl;

    char pcmout[4096];
    int current_section;
    long bytes;

    /// Corresponds to PA_SAMPLE_S16LE
    int is_big_endian = false;
    int word_size = 2;
    int is_signed = true;

    while ((bytes = ov_read(&vf, pcmout, sizeof(pcmout), is_big_endian, word_size, is_signed, &current_section)) > 0) {
        audio_data.insert(audio_data.end(), pcmout, pcmout + bytes);
    }
}

void ReMixer::Sound::LoadWAVFile(const std::filesystem::path &file_name) {
    std::ifstream file(file_name, std::ios::binary);
    if (!file.is_open()) {
        throw std::runtime_error(std::format("Failed to open WAV file: {}", file_name.c_str()));
    }

    auto read_u16 = [&](std::istream &s) -> uint16_t {
        uint8_t bytes[2];
        s.read(reinterpret_cast<char*>(bytes), 2);
        return bytes[0] | (bytes[1] << 8);
    };

    auto read_u32 = [&](std::istream &s) -> uint32_t {
        uint8_t bytes[4];
        s.read(reinterpret_cast<char*>(bytes), 4);
        return bytes[0] | (bytes[1] << 8) | (bytes[2] << 16) | (bytes[3] << 24);
    };

    char riff[4];
    file.read(riff, 4);
    if (strncmp(riff, "RIFF", 4) != 0) {
        throw std::runtime_error("Invalid WAV file: missing RIFF header.");
    }

    file.ignore(4);

    char wave[4];
    file.read(wave, 4);
    if (strncmp(wave, "WAVE", 4) != 0) {
        throw std::runtime_error("Invalid WAV file: missing WAVE header.");
    }

    uint16_t num_channels    = 0;
    uint32_t sample_rate_    = 0;
    uint16_t bits_per_sample = 0;
    std::vector<char> raw_data;

    while (!file.eof()) {
        char chunk_id[4];
        file.read(chunk_id, 4);
        if (file.eof()) break;

        uint32_t chunk_size = read_u32(file);

        if (strncmp(chunk_id, "fmt ", 4) == 0) {
            uint16_t audio_format = read_u16(file);
            num_channels = read_u16(file);
            sample_rate_ = read_u32(file);
            file.ignore(6);
            bits_per_sample = read_u16(file);

            if (audio_format != 1 || bits_per_sample != 16) {
                throw std::runtime_error("Unsupported WAV format: only PCM 16-bit is supported.");
            }

            if (chunk_size > 16) {
                file.ignore(chunk_size - 16);
            }
        }
        else if (strncmp(chunk_id, "data", 4) == 0) {
            raw_data.resize(chunk_size);
            file.read(raw_data.data(), chunk_size);
            break;
        }
        else {
            file.ignore(chunk_size);
        }
    }

    if (raw_data.empty()) {
        throw std::runtime_error("WAV file contains no audio data.");
    }

    StreamMode mode;
    if (num_channels == 1) {
        mode = StreamMode::MONO;
    } else if (num_channels == 2) {
        mode = StreamMode::STEREO;
    } else {
        throw std::runtime_error(std::format("Unsupported WAV Channel count: {}", num_channels));
    }

    std::cout << std::format("Loaded WAV file: {} Hz, {} channel(s), {} bits\n",
    sample_rate_, num_channels, bits_per_sample);

    LoadPCMBuffer(raw_data, sample_rate_, mode);
}

uint ReMixer::Sound::Channels() const {
    if (stream_mode == StreamMode::MONO) return 1;
    if (stream_mode == StreamMode::STEREO) return 2;
    return 0;
}

ReMixer::StreamMode ReMixer::Sound::GetStreamMode() const { return stream_mode;}

uint ReMixer::Sound::SampleRate() const { return sample_rate;}

float ReMixer::Sound::PlaybackLength() {
    // TODO: Factor in sample_depth properly
    float sample_depth = 2;
    float sample_count = (float)SampleCount();

    return sample_count / (SampleRate() * Channels() * sample_depth);
}

uint ReMixer::Sound::SampleCount() const {
    // TODO: Should we divide the audio buffer's real size by sample_depth?
    // Consider sample_depth = 2, each 16-bit sample would hold two elements in the char buffer.
    return audio_data.size();
}

uint ReMixer::Sound::DataSizeInBytes() const {
    return audio_data.size() * sizeof(decltype(audio_data)::value_type);
}

char ReMixer::Sound::At(uint index) const {
    return audio_data[index];
}

char &ReMixer::Sound::At(uint index) {
    return audio_data[index];
}

ReMixer::Sound ReMixer::Sound::operator+(const ReMixer::Sound &rhs) const {

    // TODO: Consider differing paramters for mixing
    // sample_rate, sample_depth, and so on.
    std::vector<char> sum;

    auto rhs_buf = rhs.audio_data;
    auto lhs_buf = this->audio_data;

    for (int i = 0; i < std::max(lhs_buf.size(), rhs_buf.size()); i++)
    {
        int a = 0;
        if (i >= lhs_buf.size() && i < rhs_buf.size()) {
            sum.push_back(rhs_buf[i]);
            continue;
        }

        if (i >= rhs_buf.size() && i < lhs_buf.size()) {
            sum.push_back(lhs_buf[i]);
            continue;
        }

        sum.push_back(mix(lhs_buf[i], rhs_buf[i]));
    }
    return Sound::FromPCMBuffer(sum);
}


void ReMixer::Sound::LoadSineWave(float length, float wavelength, float amplitude, float phase)
{
    stream_mode = StreamMode::MONO;
    // TODO: Allow sample_rate as a parameter.
    sample_rate = 44.1_kHz;

    uint samples = length * SampleRate();

    for (int i = 0; i < samples; i++)
    {
        float time = (float)i / (float)sample_rate;
        float input = (2.f*M_PI*time) - phase;
        float wave = amplitude * std::sin(input / wavelength);
        short sample = (short)(wave);
        char low_half = sample;
        char high_half = (sample >> 8);

        audio_data.push_back(low_half);
        audio_data.push_back(high_half);
    }
}


void ReMixer::Sound::LoadSawtoothWave(float length, float wavelength, float amplitude, float phase) {
    stream_mode = StreamMode::MONO;
    // TODO: Allow sample_rate as a parameter.
    sample_rate = 44.1_kHz;

    uint samples = length * SampleRate();

    for (int i = 0; i < samples; i++) {

        float time = (float) i / (float)sample_rate;
        float input = (2.f*M_PI*time) - phase;
        float coefficient = 2.f*amplitude / M_PI;
        float wave = coefficient * std::atan(std::tan( input / (2.f*wavelength)));
        short sample = (short)(wave);
        char low_half = sample;
        char high_half = (sample >> 8);

        audio_data.push_back(low_half);
        audio_data.push_back(high_half);
    }
}

void ReMixer::Sound::LoadTriangleWave(float length, float wavelength, float amplitude, float phase) {
    stream_mode = StreamMode::MONO;
    // TODO: Allow sample_rate as a parameter.
    sample_rate = 44.1_kHz;

    uint samples = length * SampleRate();

    for (int i = 0; i < samples; i++) {
        float time = (float) i / (float)sample_rate;
        float input = (2.f*M_PI*time) - phase;
        float coefficient = 2.f*amplitude / M_PI;
        float wave = coefficient * std::asin(std::sin(input / wavelength));
        short sample = (short)(wave);
        char low_half = sample;
        char high_half = (sample >> 8);

        audio_data.push_back(low_half);
        audio_data.push_back(high_half);
    }

}

void ReMixer::Sound::LoadSquareWave(float length, float wavelength, float amplitude, float phase) {
    stream_mode = StreamMode::MONO;
    // TODO: Allow sample_rate as a parameter.
    sample_rate = 44.1_kHz;

    uint samples = length * SampleRate();

    float duty = 0.5f*wavelength;

    for (int i = 0; i < samples; i++) {
        float time = (float) i / (float)sample_rate;

        float wave = -amplitude;

        if (std::fmod(time - phase, wavelength) < duty) {
            wave = amplitude;
        }

        short sample = (short)(wave);
        char low_half = sample;
        char high_half = (sample >> 8);

        audio_data.push_back(low_half);
        audio_data.push_back(high_half);
    }
}

ReMixer::Sound ReMixer::Sound::FromSineWave(float length, float wavelength, float amplitude, float phase) {
    Sound sound;
    sound.LoadSineWave(length, wavelength, amplitude, phase);
    return sound;
}

ReMixer::Sound ReMixer::Sound::FromSawtoothWave(float length, float wavelength, float amplitude, float phase) {
    Sound sound;
    sound.LoadSawtoothWave(length, wavelength, amplitude, phase);
    return sound;
}

ReMixer::Sound ReMixer::Sound::FromTriangleWave(float length, float wavelength, float amplitude, float phase) {
    Sound sound;
    sound.LoadTriangleWave(length, wavelength, amplitude, phase);
    return sound;
}

ReMixer::Sound ReMixer::Sound::FromSquareWave(float length, float wavelength, float amplitude, float phase) {
    Sound sound;
    sound.LoadSquareWave(length, wavelength, amplitude, phase);
    return sound;
}

void ReMixer::Sound::LoadNoise(float length, float amplitude) {
    stream_mode = StreamMode::MONO;
    sample_rate = 44.1_kHz;

    uint samples = length * SampleRate();

    srand(time(NULL));

    for (int i = 0; i < samples; i++) {
        float time = (float) i / (float) sample_rate;

        float wave = (rand() % 65535) - 32000;

        short sample = (short) (wave*amplitude);
        char low_half = sample;
        char high_half = (sample >> 8);

        audio_data.push_back(low_half);
        audio_data.push_back(high_half);

    }
}

ReMixer::Sound ReMixer::Sound::FromNoise(float length, float amplitude) {
    Sound sound;
    sound.LoadNoise(length, amplitude);
    return sound;
}

int mix(int a, int b) {
    return std::clamp(a + b, -128, 128);
}
